package vss

type A struct {
	IntNoMinMax int8
	IntOnlyMax int8
	IntOnlyMin int8
	IntMinMax int8
	IntMaxZero int8
	IntMinZero int8
	FloatNoMinMax float32
	FloatOnlyMax float32
	FloatOnlyMin float32
	FloatMinMax float32
	FloatMaxZero float32
	FloatMinZero float32
	FloatMaxZeroInt float32
	FloatMinZeroInt float32
}
